@@ -17,6 +17,7 @@ The return value (stdout) is a JSON object with one key, ``instance_id``.
 from __future__ import absolute_import, division, print_function, unicode_literals
 
 import os, sys, time, datetime, base64, json
+from typing import Dict, List
 
 import yaml
 
@@ -26,7 +27,8 @@ from .util import wait_for_port, validate_hostname, paginate
 from .util.cloudinit import get_user_data
 from .util.aws import (ensure_vpc, ensure_subnet, ensure_security_group, ensure_log_group, ensure_instance_profile,
                        add_tags, resolve_security_group, get_bdm, resolve_instance_id, expect_error_codes, resolve_ami,
-                       get_ondemand_price_usd, resources, clients, ARN, instance_type_completer)
+                       locate_ami, get_ondemand_price_usd, resources, clients, ARN, instance_type_completer,
+                       get_ssm_parameter, encode_tags)
 from .util.aws.dns import DNSZone, get_client_token
 from .util.aws.spot import SpotFleetBuilder
 from .util.crypto import new_ssh_key, add_ssh_host_key_to_known_hosts, ensure_ssh_key, hostkey_line
@@ -78,15 +80,25 @@ def launch(args):
         validate_hostname(args.hostname)
         assert not args.hostname.startswith("i-")
     ami_tags = dict(tag.split("=", 1) for tag in args.ami_tags or [])
-    args.ami = resolve_ami(args.ami, **ami_tags)
+    if args.ubuntu_linux_ami:
+        args.ami = locate_ami(product="com.ubuntu.cloud:server:20.04:amd64")
+    elif args.amazon_linux_ami:
+        args.ami = get_ssm_parameter("/aws/service/ami-amazon-linux-latest/amzn2-ami-hvm-x86_64-gp2")
+    else:
+        try:
+            args.ami = resolve_ami(args.ami, **ami_tags)
+        except AegeaException as e:
+            if args.ami is None and len(ami_tags) == 0 and "Could not resolve AMI" in str(e):
+                raise AegeaException("No AMI was given, and no AMIs were found in this account. "
+                                     "To use the default Ubuntu Linux LTS AMI, use --ubuntu-linux-ami. "
+                                     "To use the default Amazon Linux 2 AMI, use --amazon-linux-ami. ")
+            raise
     if args.subnet:
         subnet = resources.ec2.Subnet(args.subnet)
         vpc = resources.ec2.Vpc(subnet.vpc_id)
     else:
         vpc = ensure_vpc()
         subnet = ensure_subnet(vpc, availability_zone=args.availability_zone)
-    if not subnet.map_public_ip_on_launch:
-        raise AegeaException("Subnets without public IP mapping are not supported")
 
     if args.security_groups:
         security_groups = [resolve_security_group(sg, vpc) for sg in args.security_groups]
@@ -113,6 +125,10 @@ def launch(args):
         user_data_args["ssh_ca_keys"] = get_ssh_ca_keys(bless_config)
         user_data_args["provision_users"] = bless_config["client_config"]["remote_users"]
 
+    hkl = hostkey_line(hostnames=[], key=ssh_host_key).strip()
+    instance_tags = dict(Name=args.hostname, Owner=ARN.get_iam_username(),
+                         SSHHostPublicKeyPart1=hkl[:255], SSHHostPublicKeyPart2=hkl[255:],
+                         OwnerSSHKeyName=ssh_key_name, **dict(args.tags))
     user_data_args.update(dict(args.cloud_config_data))
     launch_spec = dict(ImageId=args.ami,
                        KeyName=ssh_key_name,
@@ -120,6 +136,7 @@ def launch(args):
                        InstanceType=args.instance_type,
                        BlockDeviceMappings=get_bdm(ebs_storage=args.storage),
                        UserData=get_user_data(**user_data_args))
+    tag_spec = dict(ResourceType="instance", Tags=encode_tags(instance_tags))
     logger.info("Launch spec user data is %i bytes long", len(launch_spec["UserData"]))
     if args.iam_role:
         instance_profile = ensure_instance_profile(args.iam_role, policies=args.iam_policies)
@@ -133,8 +150,10 @@ def launch(args):
     try:
         if args.spot:
             launch_spec["UserData"] = base64.b64encode(launch_spec["UserData"]).decode()
-            if args.duration_hours or args.cores or args.min_mem_per_core_gb:
-                spot_fleet_args = dict(launch_spec=launch_spec, client_token=args.client_token)
+            if args.duration_hours or args.cores or args.min_mem_per_core_gb or args.spot_instance_interruption_behavior != "terminate":
+                spot_fleet_args = dict(launch_spec=dict(launch_spec, TagSpecifications=[tag_spec]),
+                                       instance_interruption_behavior=args.spot_instance_interruption_behavior,
+                                       client_token=args.client_token)
                 for arg in "cores", "min_mem_per_core_gb", "spot_price", "duration_hours", "dry_run":
                     if getattr(args, arg, None):
                         spot_fleet_args[arg] = getattr(args, arg)
@@ -148,12 +167,12 @@ def launch(args):
                     class InstanceSpotFleetBuilder(SpotFleetBuilder):
                         def instance_types(self, **kwargs):
                             yield args.instance_type, 1
-                    spot_fleet_builder = InstanceSpotFleetBuilder(**spot_fleet_args)
+                    spot_fleet_builder = InstanceSpotFleetBuilder(**spot_fleet_args)  # type: SpotFleetBuilder
                 else:
                     spot_fleet_builder = SpotFleetBuilder(**spot_fleet_args)
                 logger.info("Launching {}".format(spot_fleet_builder))
                 sfr_id = spot_fleet_builder()
-                instances = []
+                instances = []  # type: List[Dict]
                 while not instances:
                     res = clients.ec2.describe_spot_fleet_instances(SpotFleetRequestId=sfr_id)
                     instances = res["ActiveInstances"]
@@ -176,7 +195,9 @@ def launch(args):
                 clients.ec2.get_waiter("spot_instance_request_fulfilled").wait(SpotInstanceRequestIds=[sir_id])
                 res = clients.ec2.describe_spot_instance_requests(SpotInstanceRequestIds=[sir_id])
                 instance = resources.ec2.Instance(res["SpotInstanceRequests"][0]["InstanceId"])
+                add_tags(instance, **instance_tags)
         else:
+            launch_spec = dict(launch_spec, TagSpecifications=[tag_spec])
             instances = resources.ec2.create_instances(MinCount=1, MaxCount=1, ClientToken=args.client_token,
                                                        DryRun=args.dry_run, **launch_spec)
             instance = instances[0]
@@ -184,18 +205,13 @@ def launch(args):
         expect_error_codes(e, "DryRunOperation")
         logger.info("Dry run succeeded")
         exit()
-    instance.wait_until_running()
-    hkl = hostkey_line(hostnames=[], key=ssh_host_key).strip()
-    tags = dict(tag.split("=", 1) for tag in args.tags)
-    add_tags(instance, Name=args.hostname, Owner=ARN.get_iam_username(),
-             SSHHostPublicKeyPart1=hkl[:255], SSHHostPublicKeyPart2=hkl[255:],
-             OwnerSSHKeyName=ssh_key_name, **tags)
-    if args.use_dns:
-        dns_zone.update(args.hostname, instance.private_dns_name)
-    while not instance.public_dns_name:
+    while instance.state["Name"] != "running":
+        logger.debug("%s %s", instance, instance.state["Name"])
         instance = resources.ec2.Instance(instance.id)
         time.sleep(1)
-    add_ssh_host_key_to_known_hosts(hostkey_line([instance.public_dns_name], ssh_host_key))
+    if args.use_dns:
+        dns_zone.update(args.hostname, instance.private_dns_name)
+    add_ssh_host_key_to_known_hosts(hostkey_line([instance.public_dns_name or instance.id], ssh_host_key))
     if args.wait_for_ssh:
         wait_for_port(instance.public_dns_name, 22)
     logger.info("Launched %s in %s using %s", instance, subnet, args.ami)
@@ -209,10 +225,13 @@ parser.add_argument("--ssh-key-name")
 parser.add_argument("--no-verify-ssh-key-pem-file", dest="verify_ssh_key_pem_file", action="store_false")
 parser.add_argument("--bless-config", default=os.environ.get("BLESS_CONFIG"),
                     help="Path to a Bless configuration file (or pass via the BLESS_CONFIG environment variable)")
-parser.add_argument("--ami", help="AMI to use for the instance. Default: " + resolve_ami.__doc__)
+parser.add_argument("--ami", help="AMI to use for the instance. Default: " + resolve_ami.__doc__)  # type: ignore
 parser.add_argument("--ami-tags", nargs="+", metavar="NAME=VALUE", help="Use the most recent AMI with these tags")
+parser.add_argument("--ubuntu-linux-ami", action="store_true", help="Use the most recent Ubuntu Linux LTS AMI")
+parser.add_argument("--amazon-linux-ami", action="store_true", help="Use the most recent Amazon Linux 2 AMI")
 parser.add_argument("--spot", action="store_true",
                     help="Launch a preemptible spot instance, which is cheaper but could be forced to shut down")
+parser.add_argument("--spot-instance-interruption-behavior", choices={"hibernate", "stop", "terminate"})
 parser.add_argument("--duration-hours", type=float, help="Terminate the spot instance after this number of hours")
 parser.add_argument("--cores", type=int, help="Minimum number of cores to request (spot fleet API)")
 parser.add_argument("--min-mem-per-core-gb", type=float)
@@ -225,7 +244,8 @@ parser.add_argument("--client-token", help="Token used to identify your instance
 parser.add_argument("--subnet")
 parser.add_argument("--availability-zone", "--az")
 parser.add_argument("--security-groups", nargs="+", metavar="SECURITY_GROUP")
-parser.add_argument("--tags", nargs="+", metavar="NAME=VALUE", help="Tags to apply to launched instances.")
+parser.add_argument("--tags", nargs="+", metavar="NAME=VALUE", type=lambda x: x.split("=", 1),
+                    help="Tags to apply to launched instances.")
 parser.add_argument("--wait-for-ssh", action="store_true",
                     help="Wait for launched instance to begin accepting SSH connections. Security groups and NACLs must permit SSH from launching host.")  # noqa
 parser.add_argument("--efs-home", action="store_true",
