@@ -1,13 +1,14 @@
 from __future__ import absolute_import, division, print_function, unicode_literals
 
-import os, sys, re, socket, time, io, gzip, logging
+import os, sys, re, socket, time, io, gzip, logging, concurrent.futures
 from functools import partial
 from datetime import datetime
 from dateutil.parser import parse as dateutil_parse
 from dateutil.relativedelta import relativedelta
+from typing import Dict, Any
 
 from .printing import GREEN
-from .compat import Repr, str
+from .compat import Repr, str, cpu_count
 
 logger = logging.getLogger(__name__)
 
@@ -63,7 +64,7 @@ class Timestamp(datetime):
     h, d, w) are supported. Negative inputs (e.g. -5m) are interpreted as relative to the current date. Other inputs
     (e.g. 2020-01-01, 15:20) are parsed using the dateutil parser.
     """
-    _precision = {}
+    _precision = {}  # type: Dict[Any, Any]
 
     def __new__(cls, t, snap=0):
         if isinstance(t, (str, bytes)) and t.isdigit():
@@ -79,7 +80,7 @@ class Timestamp(datetime):
                 snap_units = {u.rstrip("s"): 0 for u in units[units.index(list(diffs)[0]) + snap:]} if snap else {}
                 snap_units.pop("day", None)
                 snap_units.update(microsecond=0)
-                ts = datetime.now().replace(**snap_units) + relativedelta(**diffs)
+                ts = datetime.now().replace(**snap_units) + relativedelta(**diffs)  # type: ignore
                 cls._precision[ts] = snap_units
                 return ts
             return dateutil_parse(t)
@@ -128,3 +129,8 @@ def get_mkfs_command(fs_type="xfs", label="aegveph"):
         return "mkfs.ext4 -L {} -F -E lazy_itable_init,lazy_journal_init ".format(label)
     else:
         raise Exception("unknown fs_type: {}".format(fs_type))
+
+class ThreadPoolExecutor(concurrent.futures.ThreadPoolExecutor):
+    def __init__(self, **kwargs):
+        max_workers = kwargs.pop("max_workers", min(8, (cpu_count() or 1) + 4))
+        return concurrent.futures.ThreadPoolExecutor.__init__(self, max_workers=max_workers, **kwargs)
