@@ -4,6 +4,7 @@ import os, sys, json, io, gzip, time, socket, hashlib, uuid
 import requests
 from warnings import warn
 from datetime import datetime, timedelta
+from typing import List, Dict, Any
 
 import boto3, botocore.session
 from botocore.exceptions import ClientError
@@ -15,10 +16,13 @@ from ..exceptions import AegeaException
 from ..compat import str
 from . import clients, resources
 
+def get_ssm_parameter(name):
+    return clients.ssm.get_parameter(Name=name)["Parameter"]["Value"]
+
 def locate_ami(product, region=None, channel="releases", stream="released", root_store="ssd", virt="hvm"):
     """
     Examples::
-        locate_ami(product="com.ubuntu.cloud:server:16.04:amd64", channel="daily", stream="daily", region="us-west-2")
+        locate_ami(product="com.ubuntu.cloud:server:20.04:amd64", channel="daily", stream="daily", region="us-west-2")
         locate_ami(product="Amazon Linux AMI 2016.09")
     """
     if region is None:
@@ -62,15 +66,21 @@ def ensure_vpc():
         else:
             from ... import config
             logger.info("Creating VPC with CIDR %s", config.vpc.cidr[ARN.get_region()])
-            vpc = resources.ec2.create_vpc(CidrBlock=config.vpc.cidr[ARN.get_region()])
+            tags = dict(Name="aegea-vpc", managedBy="aegea")
+            tag_spec = dict(ResourceType="vpc", Tags=encode_tags(tags))
+            vpc = resources.ec2.create_vpc(CidrBlock=config.vpc.cidr[ARN.get_region()], TagSpecifications=[tag_spec])
             clients.ec2.get_waiter("vpc_available").wait(VpcIds=[vpc.id])
-            add_tags(vpc, Name=__name__)
             vpc.modify_attribute(EnableDnsSupport=dict(Value=config.vpc.enable_dns_support))
             vpc.modify_attribute(EnableDnsHostnames=dict(Value=config.vpc.enable_dns_hostnames))
-            internet_gateway = resources.ec2.create_internet_gateway()
+            tags = dict(Name="aegea-igw", managedBy="aegea")
+            tag_spec = dict(ResourceType="internet-gateway", Tags=encode_tags(tags))
+            internet_gateway = resources.ec2.create_internet_gateway(TagSpecifications=[tag_spec])
             vpc.attach_internet_gateway(InternetGatewayId=internet_gateway.id)
             for route_table in vpc.route_tables.all():
                 route_table.create_route(DestinationCidrBlock="0.0.0.0/0", GatewayId=internet_gateway.id)
+            tags = dict(Name="aegea-eigw", managedBy="aegea")
+            tag_spec = dict(ResourceType="egress-only-internet-gateway", Tags=encode_tags(tags))
+            clients.ec2.create_egress_only_internet_gateway(VpcId=vpc.id, TagSpecifications=[tag_spec])
             ensure_subnet(vpc)
     return vpc
 
@@ -93,9 +103,11 @@ def ensure_subnet(vpc, availability_zone=None):
         subnets = {}
         for az, subnet_cidr in zip(availability_zones(), subnet_cidrs):
             logger.info("Creating subnet with CIDR %s in %s, %s", subnet_cidr, vpc, az)
-            subnets[az] = resources.ec2.create_subnet(VpcId=vpc.id, CidrBlock=str(subnet_cidr), AvailabilityZone=az)
+            tags = dict(Name="aegea-subnet", managedBy="aegea")
+            tag_spec = dict(ResourceType="subnet", Tags=encode_tags(tags))
+            subnets[az] = resources.ec2.create_subnet(VpcId=vpc.id, CidrBlock=str(subnet_cidr), AvailabilityZone=az,
+                                                      TagSpecifications=[tag_spec])
             clients.ec2.get_waiter("subnet_available").wait(SubnetIds=[subnets[az].id])
-            add_tags(subnets[az], Name=__name__)
             clients.ec2.modify_subnet_attribute(SubnetId=subnets[az].id,
                                                 MapPublicIpOnLaunch=dict(Value=config.vpc.map_public_ip_on_launch))
         subnet = subnets[availability_zone] if availability_zone is not None else list(subnets.values())[0]
@@ -126,7 +138,7 @@ def resolve_security_group(name, vpc=None):
             return security_group
     raise KeyError(name)
 
-def ensure_security_group(name, vpc, tcp_ingress=[dict(port=socket.getservbyname("ssh"), cidr="0.0.0.0/0")]):
+def ensure_security_group(name, vpc, tcp_ingress=frozenset()):
     try:
         security_group = resolve_security_group(name, vpc)
     except (ClientError, KeyError):
@@ -187,6 +199,7 @@ def ensure_s3_bucket(name=None, policy=None, lifecycle=None):
     return bucket
 
 class ARN:
+    arn = partition = service = region = account_id = resource = ""
     fields = "arn partition service region account_id resource".split()
     _default_region, _default_account_id, _default_iam_username = None, None, None
 
@@ -235,7 +248,7 @@ class ARN:
 
 class IAMPolicyBuilder:
     def __init__(self, *args, **kwargs):
-        self.policy = dict(Version="2012-10-17", Statement=[])
+        self.policy = dict(Version="2012-10-17", Statement=[])  # type: Dict[str, Any]
         if args:
             if len(args) > 1 or not isinstance(args[0], dict):
                 raise AegeaException("IAMPolicyBuilder: Expected one policy document")
@@ -329,7 +342,10 @@ def ensure_iam_entity(iam_entity_name, policies, collection, constructor, **cons
     attached_policies = [policy.arn for policy in entity.attached_policies.all()]
     for policy in policies:
         if isinstance(policy, IAMPolicyBuilder):
-            entity.Policy(__name__).put(PolicyDocument=str(policy))
+            try:
+                assert entity.Policy(__name__).policy_document == policy.policy
+            except Exception:
+                entity.Policy(__name__).put(PolicyDocument=str(policy))
         else:
             policy_arn = "arn:aws:iam::aws:policy/{}".format(policy)
             if policy_arn not in attached_policies:
@@ -352,10 +368,14 @@ def ensure_instance_profile(iam_role_name, policies=frozenset()):
         instance_profile.add_role(RoleName=role.name)
     return instance_profile
 
-def encode_tags(tags):
+def encode_tags(tags, case="title"):
     if isinstance(tags, (list, tuple)):
         tags = dict(tag.split("=", 1) for tag in tags)
-    return [dict(Key=k, Value=v) for k, v in tags.items()]
+    assert case in {"title", "lower"}
+    if case == "title":
+        return [dict(Key=k, Value=v) for k, v in tags.items()]
+    elif case == "lower":
+        return [dict(key=k, value=v) for k, v in tags.items()]
 
 def decode_tags(tags):
     return {tag["Key"]: tag["Value"] for tag in tags}
@@ -379,7 +399,7 @@ def resolve_instance_id(name):
 def get_bdm(max_devices=12, ebs_storage=frozenset()):
     # Note: d2.8xl and hs1.8xl have 24 devices
     bdm = [dict(VirtualName="ephemeral" + str(i), DeviceName="xvd" + chr(ord("b") + i)) for i in range(max_devices)]
-    ebs_bdm = []
+    ebs_bdm = []  # type: List[Dict]
     for i, (mountpoint, size_gb) in enumerate(ebs_storage):
         ebs_spec = dict(Encrypted=True, DeleteOnTermination=True, VolumeType="st1", VolumeSize=int(size_gb))
         ebs_bdm.insert(0, dict(DeviceName="xvd" + chr(ord("z") - i), Ebs=ebs_spec))
@@ -455,7 +475,7 @@ def get_pricing_data(service_code, filters=None, max_cache_age_days=30):
         if cache_date < datetime.now() - timedelta(days=max_cache_age_days):
             raise Exception("Cache is too old, discard")
         with gzip.open(service_code_filename) as gz_fh:
-            with io.BufferedReader(gz_fh) as buf_fh:
+            with io.BufferedReader(gz_fh) as buf_fh:  # type: ignore
                 pricing_data = json.loads(buf_fh.read().decode())
     except Exception:
         logger.info("Fetching pricing data for %s", service_code)
