@@ -1,11 +1,12 @@
-import io, json, base64, hashlib, argparse
+import os, sys, io, json, base64, hashlib, argparse
+from typing import Dict, Any
 
 import yaml
 from botocore.exceptions import ClientError
 from botocore.paginate import Paginator
 
 from . import ARN, resources, clients, expect_error_codes, ensure_s3_bucket, ensure_iam_role, instance_storage_shellcode
-from .. import paginate, get_mkfs_command
+from .. import paginate, get_mkfs_command, logger
 from ..exceptions import AegeaException
 from ... import __version__
 
@@ -15,7 +16,7 @@ env_mgr_shellcode = """
 set -a
 if [ -f /etc/environment ]; then source /etc/environment; fi
 if [ -f /etc/default/locale ]; then source /etc/default/locale; else export LC_ALL=C.UTF-8 LANG=C.UTF-8; fi
-export AWS_DEFAULT_REGION={region}
+export AWS_DEFAULT_REGION={region} DEBIAN_FRONTEND=noninteractive
 set +a
 if [ -f /etc/profile ]; then source /etc/profile; fi
 set -euo pipefail
@@ -98,7 +99,7 @@ def get_command_and_env(args):
         shellcode += commands
 
     if args.execute:
-        bucket = ensure_s3_bucket(args.staging_s3_bucket or "aegea-batch-jobs-" + ARN.get_account_id())
+        bucket = ensure_s3_bucket(args.staging_s3_bucket)
 
         key_name = hashlib.sha256(args.execute.read()).hexdigest()
         args.execute.seek(0)
@@ -115,33 +116,24 @@ def get_command_and_env(args):
                       "curl -L '{payload_url}' > $BATCH_SCRIPT".format(payload_url=payload_url),
                       "chmod +x $BATCH_SCRIPT",
                       "$BATCH_SCRIPT"]
-    elif args.cwl:
-        ensure_dynamodb_table("aegea-batch-jobs", hash_key_name="job_id")
-        bucket = ensure_s3_bucket(args.staging_s3_bucket or "aegea-batch-jobs-" + ARN.get_account_id())
-        args.environment.append(dict(name="AEGEA_BATCH_S3_BASE_URL", value="s3://" + bucket.name))
-
-        from cwltool.main import main as cwltool_main
-        with io.BytesIO() as preprocessed_cwl:
-            if cwltool_main(["--print-pre", args.cwl], stdout=preprocessed_cwl) != 0:
-                raise AegeaException("Error while running cwltool")
-            cwl_spec = yaml.load(preprocessed_cwl.getvalue())
-            payload = base64.b64encode(preprocessed_cwl.getvalue()).decode()
-            args.environment.append(dict(name="AEGEA_BATCH_CWL_DEF_B64", value=payload))
-            payload = base64.b64encode(args.cwl_input.read()).decode()
-            args.environment.append(dict(name="AEGEA_BATCH_CWL_JOB_B64", value=payload))
-
-        for requirement in cwl_spec.get("requirements", []):
-            if requirement["class"] == "DockerRequirement":
-                # FIXME: dockerFile support: ensure_ecr_image(...)
-                # container_props["image"] = requirement["dockerPull"]
-                pass
-
+    elif args.wdl:
+        bucket = ensure_s3_bucket(args.staging_s3_bucket)
+        wdl_key_name = "{}.wdl".format(hashlib.sha256(args.wdl.read()).hexdigest())
+        args.wdl.seek(0)
+        bucket.upload_fileobj(args.wdl, wdl_key_name)
+        wdl_input = args.wdl_input.read().encode()
+        wdl_input_key_name = "{}.json".format(hashlib.sha256(wdl_input).hexdigest())
+        bucket.Object(wdl_input_key_name).put(Body=wdl_input)
         shellcode += [
-            # 'sed -i -e "s|http://archive.ubuntu.com|http://us-east-1.ec2.archive.ubuntu.com|g" /etc/apt/sources.list',
-            # "apt-get update -qq",
-            # "apt-get install -qqy --no-install-suggests --no-install-recommends --force-yes python-pip python-requests python-yaml python-lockfile python-pyparsing awscli", # noqa
-            # "pip install ruamel.yaml==0.13.4 cwltool==1.0.20161227200419 dynamoq tractorbeam",
-            "cwltool --no-container --preserve-entire-environment <(echo $AEGEA_BATCH_CWL_DEF_B64 | base64 -d) <(echo $AEGEA_BATCH_CWL_JOB_B64 | base64 -d | tractor pull) | tractor push $AEGEA_BATCH_S3_BASE_URL/$AWS_BATCH_JOB_ID | dynamoq update aegea-batch-jobs $AWS_BATCH_JOB_ID" # noqa
+            "sed -i s/archive.ubuntu.com/{}.ec2.archive.ubuntu.com/ /etc/apt/sources.list".format(ARN.get_region()),
+            "apt-get -qq update",
+            "apt-get -qq install --no-install-suggests --no-install-recommends --yes python3-{pip,setuptools,wheel}",
+            "pip3 install miniwdl awscli",
+            "cd /mnt",
+            "aws s3 cp s3://{bucket}/{key} .".format(bucket=bucket.name, key=wdl_key_name),
+            "aws s3 cp s3://{bucket}/{key} wdl_input.json".format(bucket=bucket.name, key=wdl_input_key_name),
+            "miniwdl run --dir /mnt --verbose --error-json {} --input wdl_input.json > wdl_output.json".format(wdl_key_name),  # noqa
+            "aws s3 cp wdl_output.json s3://{bucket}/wdl_output/${{AWS_BATCH_JOB_ID}}.json".format(bucket=bucket.name)
         ]
     args.command = bash_cmd_preamble + shellcode + (args.command or [])
     return args.command, args.environment
@@ -159,23 +151,35 @@ def set_ulimits(args, container_props):
             name, value = ulimit.split(":", 1)
             container_props["ulimits"].append(dict(name=name, hardLimit=int(value), softLimit=int(value)))
 
-def set_volumes(args, container_props):
+def get_volumes_and_mountpoints(args):
+    volumes, mount_points = [], []
+    if args.wdl and ["/var/run/docker.sock", "/var/run/docker.sock"] not in args.volumes:
+        args.volumes.append(["/var/run/docker.sock", "/var/run/docker.sock"])
     if args.volumes:
         for i, (host_path, guest_path) in enumerate(args.volumes):
-            container_props["volumes"].append({"host": {"sourcePath": host_path}, "name": "vol%d" % i})
-            container_props["mountPoints"].append({"sourceVolume": "vol%d" % i, "containerPath": guest_path})
+            vol_spec = {"name": "vol%d" % i}  # type: Dict[str, Any]
+            mount_spec = {"sourceVolume": "vol%d" % i, "containerPath": guest_path}
+            if host_path.startswith("fs-"):
+                fs_id, _, root_directory = host_path.partition(":")
+                vol_spec["efsVolumeConfiguration"] = dict(fileSystemId=fs_id)
+                if root_directory:
+                    vol_spec["efsVolumeConfiguration"]["rootDirectory"] = root_directory
+            else:
+                vol_spec["host"] = dict(sourcePath=host_path)
+            volumes.append(vol_spec)
+            mount_points.append(mount_spec)
+    return volumes, mount_points
 
 def ensure_job_definition(args):
     if args.ecs_image:
         args.image = get_ecr_image_uri(args.ecs_image)
-    container_props = {k: getattr(args, k) for k in ("image", "vcpus", "memory", "privileged")}
-    container_props.update(volumes=[], mountPoints=[], environment=[], command=[], resourceRequirements=[])
-    set_volumes(args, container_props)
+    container_props = {k: getattr(args, k) for k in ("image", "vcpus", "user", "privileged")}
+    container_props.update(memory=4, volumes=[], mountPoints=[], environment=[], command=[], resourceRequirements=[])
+    container_props["volumes"], container_props["mountPoints"] = get_volumes_and_mountpoints(args)
     set_ulimits(args, container_props)
     if args.gpus:
         container_props["resourceRequirements"] = [{"type": "GPU", "value": str(args.gpus)}]
-    iam_role = ensure_iam_role(args.job_role, trust=["ecs-tasks"],
-                               policies=["AmazonEC2FullAccess", "AmazonDynamoDBFullAccess", "AmazonS3FullAccess"])
+    iam_role = ensure_iam_role(args.job_role, trust=["ecs-tasks"], policies=args.default_job_role_iam_policies)
     container_props.update(jobRoleArn=iam_role.arn)
     expect_job_defn = dict(status="ACTIVE", type="container", parameters={},
                            retryStrategy={'attempts': args.retry_attempts}, containerProperties=container_props)
@@ -195,3 +199,24 @@ def ensure_job_definition(args):
                                                  type="container",
                                                  containerProperties=container_props,
                                                  retryStrategy=dict(attempts=args.retry_attempts))
+
+
+def ensure_lambda_helper():
+    awslambda = getattr(clients, "lambda")
+    try:
+        helper_desc = awslambda.get_function(FunctionName="aegea-dev-process_batch_event")
+        logger.info("Using Batch helper Lambda %s", helper_desc["Configuration"]["FunctionArn"])
+    except awslambda.exceptions.ResourceNotFoundException:
+        logger.info("Batch helper Lambda not found, installing")
+        import chalice.cli  # type: ignore
+        orig_argv = sys.argv
+        orig_wd = os.getcwd()
+        try:
+            os.chdir(os.path.join(os.path.dirname(__file__), "batch_events_lambda"))
+            sys.argv = ["chalice", "deploy", "--no-autogen-policy"]
+            chalice.cli.main()
+        except SystemExit:
+            pass
+        finally:
+            os.chdir(orig_wd)
+            sys.argv = orig_argv
