@@ -67,7 +67,7 @@ class TestAegea(unittest.TestCase):
                       dict(return_codes=[1, os.EX_SOFTWARE],
                            stderr="(UnauthorizedOperation|AccessDenied|DryRunOperation)")]
             args = []
-            if subcommand in ("ssh", "scp", "put-alarm", "put_alarm", "batch"):
+            if subcommand in ("ssh", "scp", "run", "put-alarm", "batch", "rm"):
                 args += ["--help"]
             elif subcommand == "top" and sys.version_info < (3, 5):
                 continue  # concurrent.futures.ThreadPoolExecutor thread count autotune introduced in 3.5
@@ -84,11 +84,13 @@ class TestAegea(unittest.TestCase):
             elif subcommand in ("grep", "filter"):
                 args += ["--help"] if USING_PYTHON2 else ["error", "syslog", "--start-time=-2h", "--end-time=-5m"]
                 expect.append(dict(return_codes=[os.EX_DATAERR]))
-            elif subcommand in ("launch", "build-ami"):
+            elif subcommand == "launch":
+                args += ["--no-verify-ssh-key-pem-file", "--dry-run", "test", "--ubuntu-linux-ami"]
+            elif subcommand == "build-ami":
                 args += ["--no-verify-ssh-key-pem-file", "--dry-run", "test"]
-            elif subcommand == "rm":
-                args += [resolve_ami()]
-            elif subcommand in ("secrets", "rds", "elb", "flow-logs", "deploy", "zones", "ebs", "buckets", "efs",
+            elif subcommand == "s3":
+                args += ["buckets"]
+            elif subcommand in ("secrets", "rds", "elb", "flow-logs", "deploy", "zones", "ebs", "efs",
                                 "ecr", "lambda", "configure", "sfn"):
                 args += ["ls"]
             elif subcommand == "pricing":
@@ -111,19 +113,27 @@ class TestAegea(unittest.TestCase):
     def test_dry_run_commands(self):
         unauthorized_ok = [dict(return_codes=[os.EX_OK]),
                            dict(return_codes=[1, os.EX_SOFTWARE], stderr="UnauthorizedOperation")]
-        self.call("aegea launch unittest --dry-run --storage /x=512 /y=1024",
+        self.call("aegea launch unittest --dry-run --storage /x=512 /y=1024 --ubuntu-linux-ami",
                   shell=True, expect=unauthorized_ok)
-        self.call("aegea launch unittest --dry-run --no-verify-ssh-key-pem-file",
+        self.call("aegea launch unittest --dry-run --no-verify-ssh-key-pem-file --ubuntu-linux-ami",
                   shell=True, expect=unauthorized_ok)
-        self.call("aegea launch unittest --dry-run --spot --no-verify-ssh-key-pem-file",
+        self.call("aegea launch unittest --dry-run --spot --no-verify-ssh-key-pem-file --amazon-linux-ami",
                   shell=True, expect=unauthorized_ok)
-        self.call("aegea launch unittest --dry-run --duration-hours 1 --no-verify-ssh-key-pem-file",
+        self.call("aegea launch unittest --dry-run --duration-hours 1 --no-verify-ssh-key-pem-file --amazon-linux-ami",
                   shell=True, expect=unauthorized_ok)
-        self.call("aegea launch unittest --duration 0.5 --min-mem 6 --cores 2 --dry-run --no-verify --client-token t",
+        self.call(("aegea launch unittest --duration 0.5 --min-mem 6 --cores 2 --dry-run --no-verify --client-token t "
+                   "--amazon-linux-ami"),
                   shell=True, expect=unauthorized_ok)
         self.call("aegea build-ami i --dry-run --no-verify-ssh-key-pem-file",
                   shell=True, expect=unauthorized_ok)
 
+        self.call("aegea batch submit --command pwd --dry-run", shell=True)
+        self.call("echo pwd > run.sh && aegea batch submit --execute run.sh --dry-run", shell=True)
+        self.call("aegea batch submit --wdl '{}' --dry-run".format(__file__.replace(".py", ".wdl")), shell=True)
+
+        self.call("aegea ecs run --command pwd --dry-run", shell=True)
+
+    @unittest.skipIf(sys.version_info < (3, 8), "Skipping test which is prone to rate limiting")
     def test_spot_fleet_builder(self):
         builder = SpotFleetBuilder(launch_spec={})
         self.assertEqual(set(spec["InstanceType"] for spec in builder.launch_specs()),
@@ -283,11 +293,13 @@ class TestAegea(unittest.TestCase):
         self.call("aegea secrets delete {s} --iam-role aegea.launch".format(s=secret_name), shell=True,
                   expect=unauthorized_ok)
 
-    @unittest.skipIf("CI" in os.environ, "requires elevated AWS privileges")
     def test_ensure_job_definition(self):
         from aegea.batch import submit_parser
-        jd1 = ensure_job_definition(submit_parser.parse_args(["--command", ""]))
-        jd2 = ensure_job_definition(submit_parser.parse_args(["--command", ""]))
+        args = submit_parser.parse_args(["--command", ""])
+        args.default_job_role_iam_policies = []
+        args.user, args.job_role, args.default_job_role_iam_policies = "0", "aegea.batch.worker", []
+        jd1 = ensure_job_definition(args)
+        jd2 = ensure_job_definition(args)
         self.assertEqual(jd1["jobDefinitionArn"], jd2["jobDefinitionArn"])
 
     @unittest.skipUnless("GH_AUTH" in os.environ, "requires GitHub credentials")
